import os, sys

import numpy as np
import matplotlib.pyplot as plt
import matplotlib


PID_TO_MODE = { 
    (-11, 11) : "e_e",
    (11, -11) : "e_e",                       
    (-13, 13) : "mu_mu",
    (13, -13) : "mu_mu",            
    (22, 22) : "gamma_gamma",
    (111, 22) : "pigamma",
    (22, 111) : "pigamma",            
    (321, -321) : "KK",
    (-321, 321) : "KK",
    (111, 111) : "pi0_pi0",
    (211, -211) : "pi+pi-",
    (-211, 211) : "pi+pi-",
    (None, None) : "Hadrons",
    None : "Hadrons",    
    }

MODE_TO_PID = dict((v, k) for k, v in PID_TO_MODE.items())

class ForeseeGenerator(object):
    """
    Generate LLP particles within FASER acceptance from FORESEE
    """
    
    def __init__(self, modelname, energy, mass, couplings, daughter1_pid = None, daughter2_pid = None, decay_modes = None,
                 outdir = None, path = '.', randomSeed = 12345, t0Shift = 0, zFront = -1.5, notime = False, suffix = "", 
                 tracker_decays = False, yshift = -0.065,
                 uncertainties = False, allowed_prod_modes = []):

        self.modelname = modelname
        self.energy = energy
        self.mass = mass
        self.couplings = [couplings] if isinstance(couplings, (str, int, float)) else couplings
        self.outdir = outdir
        self.path = path
        self.version = 4  # Forsee "version" 1.X
        self.seed = randomSeed
        self.zfront = zFront
        self.t0 = t0Shift + zFront
        self.notime = notime
        self.suffix = f"_{suffix}" if suffix else ""
        self.nbinsample = 1
        self.nevents = 0
        self.tracker_decays = tracker_decays
        self.uncertainties = uncertainties
        self.weightnames = []
        self.prod_modes = {}
        self.allowed_prod_modes = allowed_prod_modes

        self.distance = 480 # m
        self.radius = 0.11 # m
        self.yshift = yshift

        self.wider_decay_volume = False


        # Set decay mode ...
        
        if decay_modes:
            if daughter1_pid is not None and daughter2_pid is not None:
                sys.exit("Cannot suply both decay modes and daughter pids")
            else:
                self.decay_modes = decay_modes
        else:
            if daughter1_pid is not None and daughter2_pid is not None:
                self.daughter1_pid = daughter1_pid
                self.daughter2_pid = daughter2_pid
                decay_modes = PID_TO_MODE.get((self.daughter1_pid, self.daughter2_pid), None)
                if decay_modes is None:
                    sys.exit(f"Undefined decay to {self.daughter1_pid} + {self.daughter2_pid} for {self.modelname}")
                
                self.decay_modes = [decay_modes]

            else:
                sys.exit("Must suply either decay modes or daughter pids")

        # Set detector ...
        self.foresee = Foresee(path = self.path)
        if self.version >= 4:            
            self.foresee.rng.seed(self.seed)

        # Generate 6.5 cm high to account for translation from ATLAS to FASER coord. system
        # TODO: relax this a bit as daughters may enter even if mother doesn't
#         self.foresee.set_detector(selection="np.sqrt(x.x**2 + x.y**2)< 0.1",  
#                                   channels=[self.mode], distance=480, length=1.5 ,
#                                   luminosity=1/1000.) # 1 pb-1        

        # Generate 6.5 cm high to account for translation from ATLAS to FASER coord. system
        # Relax r to 11 cm to acount for any mismatches with calypso and for duaghter entering if mother doesn't

        if self.wider_decay_volume:
            print("*** Generating decays from just before end of magnet to end of calo ***")
            self.distance = 480 + 1.5 + 2.2 #m
            self.zfront = 2.2 #m
            length = 1.3 # m
            self.radius = 0.15 #m
            print(f"   d = {self.distance} m -> zfront = {self.zfront} m, l = {length} m, r = {self.radius} m")
        elif self.tracker_decays or "ALP" in self.modelname or all(m in ["gamma_gamma", "pigamma", "gamma", "pi0_pi0"] for m in self.decay_modes):
            # Since not relying on tracks then extend length to include spectrometer (magents + tracking stations)
            print("*** Including decays in tracker volume ***")
            length = 1.5 + 2 + 0.5 # m (to end of tracker)
            #print("*** Including decays in tracker volume and calo ***")        
            #length = 1.5 + 3.5 # m (to end of calo)            
        else:
            length = 1.5 # m



#         self.foresee.set_detector(selection="np.sqrt(x.x**2 + (x.y - 0.065)**2)< 0.11",  
#                                   channels=self.decay_modes, distance=480, length=length ,
#                                   luminosity=1/1000.) # 1 pb-1        


        # Set model ...
        self.model = Model(self.modelname, path = f"{self.path}/Models/{self.modelname}/")
        
        if self.modelname == "DarkPhoton":
            if "pigamma" in self.decay_modes:
                idx = self.decay_modes.index("pigamma")
                self.decay_modes[idx] = "pi0_gamma" 
            
            self.darkphoton()
        elif self.modelname == "ALP-W":
            self.alp_W()
        elif self.modelname == "ALP-photon":
            if "gamma_gamma" in self.decay_modes:
                idx = self.decay_modes.index("gamma_gamma")
                self.decay_modes[idx] = "gamma"             
            self.alp_photon()
        elif self.modelname == "ALP-g":
            if "gamma_gamma" in self.decay_modes:
                idx = self.decay_modes.index("gamma_gamma")
                self.decay_modes[idx] = "br_2Gamma" 

            self.alp_gluon()
        elif self.modelname == "U(1)B-L":
            self.bminusl()
        elif self.modelname == "U(1)B":
            self.decay_modes = [m.replace("_", "") for m in self.decay_modes]
            self.u1b()
        elif self.modelname == "UpPhilic":
            if "gamma_gamma" in self.decay_modes:
                idx = self.decay_modes.index("gamma_gamma")
                self.decay_modes[idx] = "gamma" 

            self.upphilic()     
        elif self.modelname == "DarkHiggs":
            self.darkhiggs()
        else:
            sys.exit(f"Unknown model {self.modelname}")


        # Setup detector along with the decay modes it is sensitive to        
        self.foresee.set_detector(selection=f"np.sqrt(x.x**2 + (x.y + {self.yshift})**2) < {self.radius}",  
                                  channels=self.decay_modes, distance=self.distance, length=length ,
                                  luminosity=1/1000.) # 1 pb-1        

        if self.modelname == "ALP-photon":
            # Position of TAN
            self.foresee.distance_prod = 140 # m

            

        # Setup decays
        self.data = self.decays()

        return

    def darkphoton(self):

        self.nbinsample = 100 # resample bins to help with asymmetric detector

        # Production modes
        self.model.add_production_2bodydecay(
            pid0 =  "111",
            pid1 = "22",
            br = "2.*0.99 * coupling**2 * pow(1.-pow(mass/self.masses('111'),2),3)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,            
            nsample = 100)
    
        self.model.add_production_2bodydecay(
            pid0 = "221",
            pid1 = "22",
            br = "2.*0.39 * coupling**2 * pow(1.-pow(mass/self.masses('221'),2),3)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET"] if self.uncertainties else "EPOSLHC",            
            energy = self.energy,
            nsample = 100)

        self.model.add_production_mixing(
            pid = "113",
            mixing = "coupling * 0.3/5. * 0.77545**2/abs(mass**2-0.77545**2+0.77545*0.147*1j)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET"] if self.uncertainties else "EPOSLHC",            
            energy = self.energy,
            )

        masses_brem = [ 
            0.01  ,  0.0126,  0.0158,  0.02  ,  0.0251,  0.0316,  0.0398,
            0.0501,  0.0631,  0.0794,  0.1   ,  0.1122,  0.1259,  0.1413,
            0.1585,  0.1778,  0.1995,  0.2239,  0.2512,  0.2818,  0.3162,
            0.3548,  0.3981,  0.4467,  0.5012,  0.5623,  0.6026,  0.631 ,
            0.6457,  0.6607,  0.6761,  0.6918,  0.7079,  0.7244,  0.7413,
            0.7586,  0.7762,  0.7943,  0.8128,  0.8318,  0.8511,  0.871 ,
            0.8913,  0.912 ,  0.9333,  0.955 ,  0.9772,  1.    ,  1.122 ,
            1.2589,  1.4125,  1.5849,  1.7783,  1.9953,  2.2387,  2.5119,
            2.8184,  3.1623,  3.9811,  5.0119,  6.3096,  7.9433, 10.    
            ]
        
        self.model.add_production_direct(
            label = "Brem",
            energy = self.energy,
            condition = ["p.pt<1", "p.pt<2", "p.pt<0.5"] if self.uncertainties else "p.pt<1",
            coupling_ref=1,
            masses = masses_brem,
            )

#         # High mass        
#         masses_dy = [ 
#             1.5849,  1.7783,  1.9953,  2.2387,  2.5119, 2.8184,  3.1623,  3.9811,  5.0119,  6.3096,  7.9433, 10.    
#             ]
#         
#         self.model.add_production_direct(
#             label = "DY",
#             energy = self.energy,
#             coupling_ref=1,
#             masses = masses_dy,
#             )

        # Production modes with order of weights for each (use nominal in case given uncertainty doesn't apply for that mode)
        if self.uncertainties:        
            self.prod_modes = {
                "111"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "221"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "113"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],                
                "Brem" : ["p.pt<1", "p.pt<1", "p.pt<1", "p.pt<1", "p.pt<2", "p.pt<0.5"]
                }
            self.weightnames = ["Nominal", "SIBYLL", "QGSJET", "Pythia8-Forward", "BremPtUp", "BremPtDown"]
        else:
            self.prod_modes = {
                "111"  : ["EPOSLHC"],
                "221"  : ["EPOSLHC"],
                "113"  : ["EPOSLHC"],
                "Brem" : ["p.pt<1"]
                }                
            self.weightnames = ["Nominal"]


        modes = ["e_e", "mu_mu", "pi+_pi-", "pi0_gamma", "pi+_pi-_pi0", "K_K"]    
        self.model.set_br_1d(
            modes = modes, # ["e_e", "mu_mu"]
            finalstates = [[11, -11], [13, -13], [211, -211], [111,22], None, [321, -321]], # [[11, -11], [13, -13]]
            filenames = [f"model/br/{m}.txt" for m in modes]    # ["model/br/e_e.txt", "model/br/mu_mu.txt"]
            )

        return #self.decays()


    def alp_W(self):
        
        nsample = 10000
        self.nbinsample = 100 # resample bins to help with smoothness

        self.model.add_production_2bodydecay(
            pid0 = "511",
            pid1 = "321",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.279**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "-511",
            pid1 = "321",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.279**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "521",
            pid1 = "321",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.279**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "-521",
            pid1 = "333",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.279**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )


        self.model.add_production_2bodydecay(
            pid0 = "531",
            pid1 = "333",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.366**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "-531",
            pid1 = "321",
            br = "2.3e4 * coupling**2 * (1-mass**2/5.366**2)**2",
            generator = ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min"] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample,
            )
        
        self.model.add_production_2bodydecay(
            pid0 = "130",
            pid1 = "111",
            br = "4.5 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.495**2)*(1-(mass-0.135)**2/0.495**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "321",
            pid1 = "211",
            br = "10.5 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.495**2)*(1-(mass-0.135)**2/0.495**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )

        self.model.add_production_2bodydecay(
            pid0 = "-321",
            pid1 = "211",
            br = "10.5 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.495**2)*(1-(mass-0.135)**2/0.495**2))",
            generator =  ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )

        # Production modes with order of weights for each (use nominal in case given uncertainty doesn't apply for that mode)
        if self.uncertainties:        
            self.prod_modes = {
                "511"  : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "-511" : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "521"  : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "-521" : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "531"  : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "-531" : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "130"  : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "321"  : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "-321" : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"]
                }
            self.weightnames = ["Nominal", "Powheg-Max", "Poweheg-Min", "SIBYLL", "QGSJET", "Pythia8-Forward"]
        else:
            self.prod_modes = {
                "511"  : ["NLO-P8"],
                "-511" : ["NLO-P8"],
                "521"  : ["NLO-P8"],
                "-521" : ["NLO-P8"],
                "531"  : ["NLO-P8"],
                "-531" : ["NLO-P8"],
                "130"  : ["EPOSLHC"],
                "321"  : ["EPOSLHC"],
                "-321" : ["EPOSLHC"]
                }
            self.weightnames = ["Nominal"]
                
        modes = ["gamma_gamma", "e_e_gamma"]
        self.model.set_br_1d(
            modes = modes, # ["gamma_gamma"],
            finalstates = [[22, 22], [11,-11,22]], # [[22, 22]],
            filenames = [f"model/br/{m}.txt" for m in modes], #["model/br/gamma_gamma.txt"]
            )   

        return #self.decays()
        

    def alp_photon(self):

        self.nbinsample = 100 # resample bins to help with smoothness
        nsample = 10000        

        def conversion_factor(mass, coupling, p0):

            # define quantities
            alpha, me, ZFe, AFe = 1./137., 0.000511, 26, 56
            a, d  = 111.*ZFe**(-1./3.) / me,  0.164 * AFe**(-2./3.)
            SMXSinIGeV2 = 13311.9696379 

            # integration boundaries theta 
            logthmin, logthmax, nlogth = -12, 0, 20
            dlogth = (logthmax-logthmin)/float(nlogth)
            
            # conversion probability 
            xs_primakoff = 0
            k = p0.e

            for ltheta in np.linspace(logthmin+0.5*dlogth,logthmax-0.5*dlogth,nlogth):
                theta = 10**ltheta
                t = mass**4 / (2.*k**2) + k**2 * theta**2
                if t<7.39*me*me: ff = a*a*t/(1.+a*a*t)
                else: ff = 1./(1.+t/d)
                xs_primakoff += alpha * ZFe**2 / 4. * ff**2 * k**4 * np.sin(theta)**3 *theta / t**2 * dlogth * np.log(10)

            prob = xs_primakoff / SMXSinIGeV2
            if np.isnan(prob): prob=0
        
            # return
            return np.sqrt(prob)

        self.model.add_production_mixing(
            label="Primakoff",
            pid = "22",
            mixing = conversion_factor,
            generator=['EPOSLHC', 'SIBYLL', 'QGSJET', 'Pythia8-Forward'] if self.uncertainties else "EPOSLHC",
            energy = self.energy
            )

#         # Remove as only relevant above 1.5 TeV or so      
#         masses = [float(x) for x in ['{:0.4e}'.format(m) for m in np.logspace(-2,0,20+1)]]
#         self.model.add_production_direct(
#             label = "Prim",
#             energy = self.energy,
#             coupling_ref = 1,
#             masses=masses,
#             )

        # Production modes with order of weights for each (use nominal in case given uncertainty doesn't apply for that mode)
        if self.uncertainties:        
            self.prod_modes = {
                "Primakoff"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"]
                }
            self.weightnames = ["Nominal", "SIBYLL", "QGSJET", "Pythia8-Forward"]
        else:
            self.prod_modes = {
                "Primakoff"  : ["EPOSLHC"]
                }
            self.weightnames = ["Nominal"]


        modes = ["gamma", "eegamma"]
        self.model.set_br_1d(
            modes = modes, # ["gamma_gamma"],
            finalstates = [[22, 22], [11,-11,22]], # [[22, 22]],
            filenames = [f"model/br/{m}.txt" for m in modes], #["model/br/gamma_gamma.txt"]
            )   
        
        return #self.decays()

    def alp_gluon(self):

        self.nbinsample = 100 # resample bins to help with smoothness
        nsample = 10000

        self.model.add_production_2bodydecay(
            pid0 = "511",
            pid1 = "130",
            br = "76 * coupling**2 * (1-(mass/5.279)**2)**2",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample, 
            ) 
        
        self.model.add_production_2bodydecay(
            pid0 = "-511",
            pid1 = "130",
            br = "76 * coupling**2 * (1-(mass/5.279)**2)**2",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample, 
            ) 
        
        self.model.add_production_2bodydecay(
            pid0 = "521",
            pid1 = "321",
            br = "76 * coupling**2 * (1-(mass/5.279)**2)**2",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample, 
            ) 
        
        self.model.add_production_2bodydecay(
            pid0 = "-521",
            pid1 = "-321",
            br = "76 * coupling**2 * (1-(mass/5.279)**2)**2",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else "NLO-P8",
            energy = self.energy,
            nsample = nsample, 
            )

        self.model.add_production_mixing(
            pid = "111",
            mixing = "(3.671) * coupling * (-0.180 * mass**2)/(mass**2 - 0.1349**2)",
            generator = ['EPOSLHC', 'SIBYLL', 'QGSJET', 'Pythia8-Forward'] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            )
        
        self.model.add_production_mixing(
            pid = "221",
            mixing = "(3.671) * coupling * (-0.3596 * mass**2 + 0.0021)/(mass**2 - 0.5478**2)",
            generator = ['EPOSLHC', 'SIBYLL', 'QGSJET', 'Pythia8-Forward'] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            )
        
        self.model.add_production_mixing(
            pid = "331",
            mixing = "(3.671) * coupling * (-0.3362 * mass**2 + 0.0092)/(mass**2 - 0.9574**2)",
            generator = ['EPOSLHC', 'SIBYLL', 'QGSJET', 'Pythia8-Forward'] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            )

        # Production modes with order of weights for each (use nominal in case given uncertainty doesn't apply for that mode)
        if self.uncertainties:        
            self.prod_modes = {
                "511"  : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "-511" : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "521"  : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "-521" : ["NLO-P8", "NLO-P8-Max", "NLO-P8-Min", "NLO-P8", "NLO-P8", "NLO-P8"],
                "111"  : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],                
                "221"  : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "331"  : ["EPOSLHC", "EPOSLHC", "EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"]
                }
            self.weightnames = ["Nominal", "Powheg-Max", "Poweheg-Min", "SIBYLL", "QGSJET", "Pythia8-Forward"]
        else:
            self.prod_modes = {
                "511"  : ["NLO-P8"],
                "-511" : ["NLO-P8"],
                "521"  : ["NLO-P8"],
                "-521" : ["NLO-P8"],
                "111"  : ["EPOSLHC"],
                "221"  : ["EPOSLHC"],
                "331"  : ["EPOSLHC"]
                }
            self.weightnames = ["Nominal"]


        modes =  ["br_2Gamma", "br_3Pi0", "br_Pi+Pi-Gamma", "br_Pi+Pi-Pi0", "br_2PiEta"]     
        self.model.set_br_1d(
            modes=modes,
            finalstates=[[22,22], [111,111,111], [211,-211,22], [211,-211, 111], [211,-211,221]],
            filenames=[f"model/br/{m}.dat" for m in modes],
            )

        return #self.decays()


    def bminusl(self):

        self.nbinsample = 100 

        self.model.add_production_2bodydecay(
            pid0 =  "111",
            pid1 = "22",
            br = "2.*0.99 * (coupling/0.303)**2 * pow(1.-pow(mass/self.masses('111'),2),3)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = 100)
    
        self.model.add_production_2bodydecay(
            pid0 = "221",
            pid1 = "22",
            br = "2.*0.25*0.39 * (coupling/0.303)**2 * pow(1.-pow(mass/self.masses('221'),2),3)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = 100)


        masses_brem = [ 
            0.01  ,  0.0126,  0.0158,  0.02  ,  0.0251,  0.0316,  0.0398,
            0.0501,  0.0631,  0.0794,  0.1   ,  0.1122,  0.1259,  0.1413,
            0.1585,  0.1778,  0.1995,  0.2239,  0.2512,  0.2818,  0.3162,
            0.3548,  0.3981,  0.4467,  0.5012,  0.5623,  0.6026,  0.631 ,
            0.6457,  0.6607,  0.6761,  0.6918,  0.7079,  0.7244,  0.7413,
            0.7586,  0.7762,  0.7943,  0.8128,  0.8318,  0.8511,  0.871 ,
            0.8913,  0.912 ,  0.9333,  0.955 ,  0.9772,  1.    ,  1.122 ,
            1.2589,  1.4125,  1.5849,  1.7783,  1.9953,  2.2387,  2.5119,
            2.8184,  3.1623,  3.9811,  5.0119,  6.3096,  7.9433, 10.    
            ]
        
        self.model.add_production_direct(
            label = "Brem",
            energy = self.energy,
            condition = ["p.pt<1", "p.pt<2", "p.pt<0.5"] if self.uncertainties else "p.pt<1",            
            coupling_ref=0.303, 
            masses = masses_brem,
            )

        if self.uncertainties:        
            self.prod_modes = {
                "111"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "221"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "Brem" : ["p.pt<1", "p.pt<1", "p.pt<1", "p.pt<1", "p.pt<2", "p.pt<0.5"]                
                }
            self.weightnames = ["Nominal", "SIBYLL", "QGSJET", "Pythia8-Forward", "BremPtUp", "BremPtDown"]
        else:
            self.prod_modes = {
                "111"  : ["EPOSLHC"],
                "221"  : ["EPOSLHC"],
                "Brem" : ["p.pt<1"],
                }
            self.weightnames = ["Nominal"]


        modes = ["e_e", "mu_mu", "nu_nu", "Hadrons"]
        self.model.set_br_1d(
            modes = modes, 
            finalstates = [[11, -11], [13, -13], [12, -12], None],
            filenames=[f"model/br/{m}.txt" for m in modes],
            )
        
        return #self.decays()


    def u1b(self):

        self.nbinsample = 100 
        nsample = 10000

        self.model.add_production_2bodydecay(
            pid0 = "111",
            pid1 = "22",
            br = "2*0.99 * (coupling/0.3)**2 * pow(1.-pow(mass/self.masses('111'),2),3)*np.abs((1-mass**2/0.78265**2 - 0.00849/0.78265*1j)**(-1))**2",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample
            )
        
        self.model.add_production_2bodydecay(
            pid0 = "221",
            pid1 = "22",
            br = "1./16*0.3941 * (coupling/0.3)**2 * pow(1.-pow(mass/self.masses('221'),2),3)*np.abs((1-mass**2/0.78265**2 - 0.00849/0.78265*1j)**(-1)+(1-mass**2/1.019461**2-0.004249/1.019461*1j)**(-1))**2",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample
            )

        self.model.add_production_2bodydecay(
            pid0 = "331",
            pid1 = "22",
            br = "1./49.*0.02307 * (coupling/0.3)**2 * pow(1.-pow(mass/self.masses('331'),2),3) * np.abs((1-mass**2/0.78265**2 - 0.00849/0.78265*1j)**(-1) - 2*(1-mass**2/1.019461**2-0.004249/1.019461*1j)**(-1))**2",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample
            )

        self.model.add_production_mixing(
            pid = "223",
            mixing = "coupling * 4./17. * 0.78265**2/abs(mass**2-0.78265**2+0.78265*0.00849*1j)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            )
        
        self. model.add_production_mixing(
            pid = "333",
            mixing = "coupling /12.88 * 1.019461**2/abs(mass**2-1.019461**2+1.019461*0.004249*1j)",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            )
        
        masses_brem = [ 
            0.01  ,  0.0126,  0.0158,  0.02  ,  0.0251,  0.0316,  0.0398,
            0.0501,  0.0631,  0.0794,  0.1   ,  0.1122,  0.1259,  0.1413,
            0.1585,  0.1778,  0.1995,  0.2239,  0.2512,  0.2818,  0.3162,
            0.3548,  0.3981,  0.4467,  0.5012,  0.5623,  0.6026,  0.631 ,
            0.6457,  0.6607,  0.6761,  0.6918,  0.7079,  0.7244,  0.7413,
            0.7586,  0.7762,  0.7943,  0.8128,  0.8318,  0.8511,  0.871 ,
            0.8913,  0.912 ,  0.9333,  0.955 ,  0.9772,  1.    ,  1.122 ,
            1.2589,  1.4125,  1.5849,  1.7783,  1.9953,  2.2387,  2.5119,
            2.8184,  3.1623,  3.9811,  5.0119,  6.3096,  7.9433, 10.    
            ]
        
        self.model.add_production_direct(
            label = "Brem",
            energy = self.energy,
            condition = ["p.pt<1", "p.pt<2", "p.pt<0.5"] if self.uncertainties else "p.pt<1",   
            coupling_ref = 0.303,          
            masses = masses_brem,
            )

        if self.uncertainties:        
            self.prod_modes = {
                "111"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "221"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "331"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "223"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],
                "333"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward", "EPOSLHC", "EPOSLHC"],                 
                "Brem" : ["p.pt<1", "p.pt<1", "p.pt<1", "p.pt<1", "p.pt<2", "p.pt<0.5"]                
                }
            self.weightnames = ["Nominal", "SIBYLL", "QGSJET", "Pythia8-Forward", "BremPtUp", "BremPtDown"]
        else:
            self.prod_modes = {
                "111"  : ["EPOSLHC"],
                "221"  : ["EPOSLHC"],
                "331"  : ["EPOSLHC"],
                "323"  : ["EPOSLHC"],
                "333"  : ["EPOSLHC"],                
                "Brem" : ["p.pt<1"]
                }
            self.weightnames = ["Nominal"]
        
        modes = ["pigamma", "3pi", "KK", "ee", "mumu"]
        self.model.set_br_1d(
            modes = modes,
            finalstates = [[111,22], [211,-211,111], [321,-321], [11,-11], [13,-13]],
            filenames=[f"model/br/{m}.txt" for m in modes],
            )
        
        return #self.decays()

    def upphilic(self):

        self.nbinsample = 100
        nsample = 10000

        self.model.add_production_2bodydecay(
            pid0 = "221",
            pid1 = "111",
            br = "1.26e5 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.547**2)*(1-(mass-0.135)**2/0.547**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )
        
        self.model.add_production_2bodydecay(
            pid0 = "331",
            pid1 = "111",
            br = "273. * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.957**2)*(1-(mass-0.135)**2/0.957**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )
        
        self.model.add_production_2bodydecay(
            pid0 = "321",
            pid1 = "211",
            br = "7.42 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.49368**2)*(1-(mass-0.135)**2/0.49368**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )
        
        self.model.add_production_2bodydecay(
            pid0 = "-321",
            pid1 = "-211",
            br = "7.42 * coupling**2 * np.sqrt((1-(mass+0.135)**2/0.49368**2)*(1-(mass-0.135)**2/0.49368**2))",
            generator = ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"] if self.uncertainties else "EPOSLHC",
            energy = self.energy,
            nsample = nsample,
            )

        if self.uncertainties:        
            self.prod_modes = {
                "221"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "331"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "321"  : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],
                "-321" : ["EPOSLHC", "SIBYLL", "QGSJET", "Pythia8-Forward"],                               
                }
            self.weightnames = ["Nominal", "SIBYLL", "QGSJET", "Pythia8-Forward"]
        else:
            self.prod_modes = {
                "221"  : ["EPOSLHC"],
                "331"  : ["EPOSLHC"],
                "321"  : ["EPOSLHC"],
                "-321" : ["EPOSLHC"],                
                }
            self.weightnames = ["Nominal"]


        modes = ["gamma", "pi+_pi-", "pi0_pi0"]
        self.model.set_br_1d(
            modes = modes, 
            finalstates = [[22,22], [211,-211], [111, 111]],
            filenames=[f"model/br/{m}.txt" for m in modes],
            )
        
        return #self.decays()

    def darkhiggs(self):

        self.nbinsample = 100 # resample bins to help with asymmetric detector

        self.model.add_production_2bodydecay(
            pid0 = "511",
            pid1 = "130",
            br = "5.6 * coupling**2 * pow(1.-pow(mass/5.279,2),2)",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else ["NLO-P8"],
            energy = self.energy,
            nsample = 10000,
        )
        
        self.model.add_production_2bodydecay(
            pid0 = "-511",
            pid1 = "130",
            br = "5.6 * coupling**2 * pow(1.-pow(mass/5.279,2),2)",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else ["NLO-P8"],
            energy = self.energy,
            nsample = 10000,
        )
        
        self.model.add_production_2bodydecay(
            pid0 = "521",
            pid1 = "321",
            br = "5.6 * coupling**2 * pow(1.-pow(mass/5.279,2),2)",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else ["NLO-P8"],
            energy = self.energy,
            nsample = 10000,
        )
        
        self.model.add_production_2bodydecay(
            pid0 = "-521",
            pid1 = "321",
            br = "5.6 * coupling**2 * pow(1.-pow(mass/5.279,2),2)",
            generator = ['NLO-P8','NLO-P8-Max', 'NLO-P8-Min'] if self.uncertainties else ["NLO-P8"],
            energy = self.energy,
            nsample = 10000,
        )        

        if self.uncertainties:        
            self.prod_modes = {
                '511':    ['NLO-P8', 'NLO-P8-max', 'NLO-P8-min'],
                '-511':   ['NLO-P8', 'NLO-P8-max', 'NLO-P8-min'],
                '521':    ['NLO-P8', 'NLO-P8-max', 'NLO-P8-min'],
                '-521':   ['NLO-P8', 'NLO-P8-max', 'NLO-P8-min'],
                }
            self.weightnames = ["Nominal", "Powheg-Max", "Poweheg-Min"]
        else:
            self.prod_modes = {
                "511"  : ['NLO-P8'],
                "-511"  : ['NLO-P8'],
                "521"  : ['NLO-P8'],
                "-521" : ['NLO-P8'],                
                }
            self.weightnames = ["Nominal"]


        modes = ["e_e", "mu_mu", "K_K", "pi_pi"]
        self.model.set_br_1d(
            modes=modes,
            finalstates=[[11,-11], [13,-13], [321,-321], [211,-211]],
            filenames=["model/br/"+mode+".txt" for mode in modes],
        )

        return

    def decays(self):
        # Set up liftime and BRs

        self.model.set_ctau_1d(
            filename=f"model/ctau.txt", 
            coupling_ref=1
            )
            
        # Get LLP spectrum
        self.foresee.set_model(model=self.model)
        # This is just a reference coupling 
        plt = self.foresee.get_llp_spectrum(self.mass, coupling=1, do_plot=True)  
        plt.savefig(f"{self.modelname}_m{self.mass}.png")

        def flatten(l):
            return [i for sublist in l for i in sublist]

        # Get list of events within detector
        output = self.foresee.get_events(mass=self.mass, energy=self.energy, couplings=self.couplings)        

        if self.version >= 3:
            coups, ctaus, nevents, momenta, weights = output
            if self.version >= 4:
                fmomenta = momenta
                fweights = weights
            else:
                fmomenta  = flatten(momenta)
                fweights  = flatten(weights)            

            fenergies = [p.e for p in fmomenta] # Now in MeV ?
            fthetas   = [(p.pt/p.pz if p.pz else 0) for p in fmomenta]            

            #self.plot(fthetas, fenergies, fweights)

            self.nevents = nevents

            # Return energy, theta and weights
            return [fenergies, fthetas, fweights]             
        else:
            coups, ctaus, nsigs, energies, weights, thetas = output
            #self.plot(flatten(thetas), flatten(energies), flatten(weights))

            # Return energy (converting to MeV), theta and weights
            return [[e*1000 for e in flatten(energies)], flatten(thetas), flatten(weights)] 

    def plot(self, thetas, energies, weights):
        # Plot the results in Forsee format

        t = np.array(thetas)
        p = np.sqrt(np.array(energies)**2 - self.mass**2)

        prange=[[-6, 0, 120],[ 0, 5, 50]]
        tmin, tmax, tnum = prange[0]
        pmin, pmax, pnum = prange[1]
        t_edges = np.logspace(tmin, tmax, num=tnum+1)
        p_edges = np.logspace(pmin, pmax, num=pnum+1)  

        ticks = np.array([[np.linspace(10**(j),10**(j+1),9)] for j in range(-7,6)]).flatten()
        ticks = [np.log10(x) for x in ticks]
        ticklabels = np.array([[r"$10^{"+str(j)+"}$","","","","","","","",""] for j in range(-7,6)]).flatten()
        matplotlib.rcParams.update({'font.size': 15})

        fig = plt.figure(figsize=(8,5.5))
        ax = plt.subplot(1,1,1)
        h=ax.hist2d(x=np.log10(t),y=np.log10(p),weights=weights,
                    bins=[tnum,pnum],range=[[tmin,tmax],[pmin,pmax]],
                    norm=matplotlib.colors.LogNorm(), cmap="hsv",
                    )
        
        fig.colorbar(h[3], ax=ax)
        ax.set_xlabel(r"angle wrt. beam axis $\theta$ [rad]")
        ax.set_ylabel(r"momentum $p$ [GeV]")
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticklabels)
        ax.set_yticks(ticks)
        ax.set_yticklabels(ticklabels)
        ax.set_xlim(tmin, tmax)
        ax.set_ylim(pmin, pmax)
        plt.savefig(f"{self.modelname}_m{self.mass}_acc.png")


    def events(self, lumi, f):
        f.write(f"{self.mass*1000} {self.couplings[0]} {self.nevents[0] * 1000 * lumi}\n")
        print(f"--> {self.mass*1000} {self.couplings[0]} {self.nevents[0] * 1000 * lumi}")        
        return
        

    def write(self):
        # Write LLP results to a file
        
        energies, thetas, weights = self.data

        if self.outdir is None:
            self.outdir = f"{self.foresee.dirpath}/Models/{self.modelname}/model/events"

        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)

        try:
            if len(self.couplings) == 1:
                filename = f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_c{self.couplings[0]}to_{self.daughter1_pid}_{self.daughter2_pid}.npy"
            else:
                filename = f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_to_{self.daughter1_pid}_{self.daughter2_pid}.npy"
        except AttributeError:
            if len(self.couplings) == 1:
                filename = f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_c{self.couplings[0]}to_{'+'.join(self.decay_modes)}.npy"
            else:
                filename = f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_to_{'+'.join(self.decay_modes)}.npy"
            
                
        print(f"Generated {len(thetas)} events")
        print(f"save data to file: {filename}")
        np.save(filename,[energies,thetas, weights])

        cfgname = filename.replace(".npy", ".cfg")
        print(f"save config to file: {cfgname}")
        with open(cfgname, "w") as f:
            f.write(" ".join(sys.argv))

        return

    def write_hepmc(self, nevents):

        if self.outdir is None:
            self.outdir = "model/events/"
        elif not os.path.exists(self.outdir):
            os.mkdir(self.outdir)

        try:
            filename =  f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_c{self.couplings[0]:.1e}to_{self.daughter1_pid}_{self.daughter2_pid}{self.suffix}.hepmc"
        except AttributeError:
            filename =  f"{self.outdir}/events_{self.energy}TeV_m{self.mass}GeV_c{self.couplings[0]:.1e}to_{'+'.join(self.decay_modes)}{self.suffix}.hepmc"            

        if self.version >= 4:
            _, weights, _ = self.foresee.write_events(self.mass, self.couplings[0], self.energy, filename, nevents,
                                                      zfront = self.zfront, # seed = self.seed, # seed now set differently
                                                      modes = ({k: v for k,v in self.prod_modes.items()
                                                                if (not self.allowed_prod_modes or k in self.allowed_prod_modes)}),
                                                      weightnames = self.weightnames,
                                                      notime = self.notime, t0 = self.t0, nsample = self.nbinsample,
                                                      return_data = True)
        else:
            _, weights, _ = self.foresee.write_events(self.mass, self.couplings[0], self.energy, filename, nevents,
                                                      zfront = self.zfront, seed = self.seed,
                                                      modes = ({k: v for k,v in self.prod_modes.items()
                                                                if (not self.allowed_prod_modes or k in self.allowed_prod_modes)}),
                                                      weightnames = self.weightnames,
                                                      notime = self.notime, t0 = self.t0, nsample = self.nbinsample,
                                                      return_data = True)


        cfgname = f"{self.foresee.dirpath}/Models/{self.modelname}/" + filename.replace(".hepmc", ".cfg")
        print(f"save config to file: {cfgname}")
        with open(cfgname, "w") as f:
            f.write(" ".join(sys.argv))

        filename = f"{self.foresee.dirpath}/Models/{self.modelname}/{filename}"

        return filename, weights.T[0]


def setup_foresee(path):

    if path is None:
        return

    # Add foresee to python path
    path = os.path.expandvars(os.path.expanduser(path))
    os.sys.path.append(f"{path}/FORESEE/src")

    # Symlink foresee files/Models dirs to current dir
    #if not os.path.exists("files"):
    #    os.symlink(os.path.expandvars(f"{path}/FORESEE/files"), "files")
    #if not os.path.exists("Models"):
    #    os.symlink(os.path.expandvars(f"{path}/FORESEE/Models"), "files")        

    # Install scikit-hep if needed.

    try:
        from skhep.math.vectors import LorentzVector, Vector3D
    except ModuleNotFoundError:
        os.system("pip install scikit-hep --user")
        try:
            from skhep.math.vectors import LorentzVector, Vector3D
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Unable to find skhep.  Please install the scikit-hep package")
        
    return

def add_to_python_path(path):
    if path in sys.path: return
    path = os.path.expandvars(os.path.expanduser(path))
    os.sys.path.append(path)
    return

def parse_couplings(data, write_hepMC = False):    

    if write_hepMC:
        if len(data) == 1:
            couplings = float(data[0])
        else:
            sys.exit("Only a single coupling allowed when writing HEPMC events")

    try:
        couplings = [float(d) for d in data]
    except ValueError:
        try:
            couplings = np.logspace(*eval(data[0]))
        except:
            sys.exit("Unable to parse couplings")

    return couplings

def production(args):

    import json
    import subprocess as proc

    args.hepmc = True

    with open(args.model) as f:
        data = json.load(f)

    xsect = {}
    nrun = data["nrun"]

    if args.outdir:
        outdir = args.outdir + "/" + data["name"]
        if not os.path.exists(outdir):
            os.mkdir(outdir)
    else:
        outdir = "."
    
    for coup, masses in data["samples"].items():        
        couplings = parse_couplings([coup], args.hepmc)

        for m in masses:
            if "modes" in data:
                if isinstance(data['modes'], str):
                    data['modes'] = [data['modes']]
                print(f">>> {nrun}: {data['model']} ({m} GeV) -> {','.join(data['modes'])} with couplings = {couplings[0]:.2e} @ Ecom = {args.Ecom} TeV")
            else:
                print(f">>> {nrun}: {data['model']} ({m} GeV) -> {data['pid1']} + {data['pid2']} with couplings = {couplings[0]:.2e} @ Ecom = {args.Ecom} TeV") 

            f = ForeseeGenerator(data["model"], args.Ecom, m, couplings, data.get("pid1", None), data.get("pid2", None),
                                 data.get("modes", None), path = args.path, randomSeed = args.randomSeed, t0Shift = args.t0, 
                                 zFront = args.zfront, yshift = data.get("yshift", args.yshift),
                                 notime = args.notime, suffix = data["name"], tracker_decays = data.get("tracker_decays", False),
                                 uncertainties = data.get("uncertainties", False), allowed_prod_modes=args.production)

            with open("events.txt", "a") as fout:
                f.events(40, fout)
                #continue

            if args.hepmc:
                filename, weights = f.write_hepmc(data["nevents"])
            else:
                f.write()

            if args.outdir:
                p = proc.run(["cp", filename, outdir + "/" + filename.split("/")[-1].replace(".hepmc", f"{nrun}.hepmc")])
                if not p.returncode:
                    proc.run(["rm", filename])

            xsect[nrun] = {
                "Mass (GeV)" : m,
                "Coupling" : f"{couplings[0]:e}",
                "Cross-section (pb)" : sum(weights), # / data["nevents"],
                "ECom (TeV)" : float(args.Ecom),
                "NEvents" : float(data["nevents"]),
                "WtNames" : f.weightnames,
                }

            if "modes" in data:
                xsect[nrun]["Description"] = f"{data['model']} ({m} GeV) -> {','.join(data['modes'])} with couplings = {couplings[0]:.2e} @ Ecom = {args.Ecom} TeV",                            
            else:
                xsect[nrun]["Description"] = f"{data['model']} ({m} GeV) -> {data['pid1']} + {data['pid2']} with couplings = {couplings[0]:.2e} @ Ecom = {args.Ecom} TeV",            

            print("X-sect", xsect)
            with open (outdir + "/cross-sections.json", "w") as f:
                json.dump(xsect, f, indent = 4)
            

            nrun += 1

  

    return

if __name__ == "__main__":

    import argparse, sys
    
    parser = argparse.ArgumentParser(description="Run FORSEE generation")
    parser.add_argument("model", help = "Name of foresee model")
    parser.add_argument("--mass", "-m", required = False, type = float, help = "Mass of mother [GeV]")
    parser.add_argument("--couplings", "-c", required = False, nargs = "+", help = "Couplings of mother (either single/mulitple values or tuple to pass to np.logspace)")
    parser.add_argument("--pid1", default = None, required = False, type = int, help = "PID of daughter 1")
    parser.add_argument("--pid2", default = None, type = int, help = "PID of daughter 2 (if not set then will be -PID1)")
    parser.add_argument("--modes", default = None, required = False, nargs = "*", help = "Space-separated list of foresee decay modes")   
    parser.add_argument("--Ecom", default = "13.6", help = "Center of mass energy [TeV]")
    parser.add_argument("--outdir", "-o", default = None, help = "Output path")    
    parser.add_argument("--path", default = ".", help = "Path to foresee installation")
    parser.add_argument("--hepmc", action = "store_true", help = "Write HepMC events")
    parser.add_argument("--nevents", "-n", default = 10, type = int, help = "Number of HepMC events ")
    parser.add_argument("--randomSeed", "-s", default = 1234, type = int, help = "Random seed for HepMC generation")
    parser.add_argument("--t0", "-t", default = 0, type = int, help = "Time offset for start of decay volume")
    parser.add_argument("--zfront", "-z", default = -1.5, help = "Minimum FASER z coordinate for decays (in meters)")
    parser.add_argument("--notime", action = "store_true", help = "Set all vertex times to 0 rather than calculating from start of decay volume")
    parser.add_argument("--suffix", default = "", help = "Filename suffix")
    parser.add_argument("--tracker", default = False, action = "store_true", help = "Include decays in the tracker")    
    parser.add_argument("--uncertainties", "-u", default = False, action = "store_true", help = "Store weights for uncertainties")    
    parser.add_argument("--yshift", "-y", default = -0.065, 
                        help = "Shift due to crossing angle (minus any detector offset) in m with negative being down")
    parser.add_argument("--production", default = None, required = False, nargs = "*", help = "Space-separated list of foresee production modes")   
    
    args = parser.parse_args()

    add_to_python_path(f"{args.path}/src")

    from foresee import Foresee, Model, Utility

    # Create PIDs
    if args.pid2 is None and args.pid1 is not None:
        args.pid2 = -args.pid1

    if args.model.endswith(".json"):
        production(args)
    else:    
    
        couplings = parse_couplings(args.couplings, args.hepmc)

        print(f"Generating {args.model} events at Ecom = {args.Ecom}") 
        print(f"   mother mass = {args.mass} GeV")
        print(f"   decay = {args.pid1} {args.pid2}")
        print(f"   couplings = {couplings}")    
        
        f = ForeseeGenerator(args.model, args.Ecom, args.mass, couplings, args.pid1, args.pid2, args.modes,
                             outdir = args.outdir, path = args.path, randomSeed = args.randomSeed, t0Shift = args.t0, zFront = args.zfront,
                             notime = args.notime, suffix = args.suffix, tracker_decays = args.tracker, yshift = args.yshift,
                             uncertainties = args.uncertainties, allowed_prod_modes=args.production)
        
        if args.hepmc:
            f.write_hepmc(args.nevents)
        else:
            f.write()
    
        

